-- TODO:
-- Search may return a truncated list of entries. Do subsearches afterwards for completeness.

-----------------------------------------------------------------------------
--                                                                         --
--  PoliteWhisper.lua - Main code for PoliteWhisper WoW add-on.            --
--                                                                         --
--  Gre7g Luterman <gre7g.luterman@gmail.com>                              --
--  Created 3/20/2007                                                      --
--                                                                         --
-----------------------------------------------------------------------------

-- $HeadURL: svn://svn.kyoht.com/wow/trunk/PoliteWhisper/PoliteWhisper.lua $
-- $LastChangedRevision: 11 $
-- $LastChangedDate: 2007-04-11 08:29:20 -0600 (Wed, 11 Apr 2007) $
-- $LastChangedBy: gre7g $

-- Constants:
PoliteWhisper_WhoDelay            = 5;   -- don't /who more than once every 5s
PoliteWhisper_WhoTimeout          = 10;  -- expect /who to respond within 10s
PoliteWhisper_AFKPesterTimeout    = 300; -- OK to ask AFK's again in 5m
PoliteWhisper_NoRespPesterTimeout = 600; -- OK to ask no responders again in 10
PoliteWhisper_ResetAllTimeout     = 3600; -- 1hr to reset all
PoliteWhisper_MaxListEntries      = 12;  -- 12 items on scroll list maximum
PoliteWhisper_MinListEntries      = 8;   -- 8 items on scroll list minimum
PoliteWhisper_IgnoreListEntries   = 18;  -- 18 items on ignore scroll list
PoliteWhisper_History             = 4;   -- Keep the last 4 messages from users
PoliteWhisper_MaxSliderHeight     = 300; -- Tall slider.
PoliteWhisper_MinSliderHeight     = 200; -- Short slider.
PoliteWhisper_AskMessage          = "Sorry to pester, " ..
    "but you would be interested in joining us on a run through %s?";
PoliteWhisper_DeclineMessage      = "No problem. Sorry to have bothered you.";
PoliteWhisper_InviteMessage       = "Hey, I found a level %d %s who's " ..
    "willing to join us. Could you please invite %s? Thanks!";
PoliteWhisper_GroupText           = "So far, we have";
PoliteWhisper_ScrollList          = {};  -- Table of scrolling items
PoliteWhisper_ChatList            = {};  -- Table of scrolling chat text
PoliteWhisper_ClassEnables        = nil; -- Table of CheckButton's
PoliteWhisper_Dungeons  =
{
    "Ahn'Qiraj",               "Alterac Valley",
    "Arathi Basin",            "Auchenai Crypts",
    "Battle of Mount Hyjal",   "Black Temple",
    "Blackfathom Deeps",       "Blackrock Depths",
    "Blackrock Mountain",      "Blackrock Spire",
    "Blackwing Lair",          "Dire Maul",
    "Eye of the Storm",        "Gnomeregan",
    "Gruul's Lair",            "Hellfire Ramparts",
    "Karazhan",                "Magtheridon's Lair",
    "Mana-Tombs",              "Maraudon",
    "Molten Core",             "Naxxramas",
    "Old Hillsbrad Foothills", "Onyxia's Lair",
    "Ragefire Chasm",          "Razorfen Downs",
    "Razorfen Kraul",          "Ruins of Ahn'Qiraj",
    "Scarlet Monastery",       "Scholomance",
    "Sethekk Halls",           "Shadow Labyrinth",
    "Shadowfang Keep",         "Stratholme",
    "Sunken Temple",           "The Arcatraz",
    "The Black Morass",        "The Blood Furnaces",
    "The Botanica",            "The Deadmines",
    "The Mechanar",            "The Shattered Halls",
    "The Slave Pens",          "The Steamvault",
    "The Stockade",            "The Temple of Atal'Hakkar",
    "The Underbog",            "Uldaman",
    "Wailing Caverns",         "Warsong Gulch",
    "Zul'Farrak",              "Zul'Gurub"
};

-- Globals:
PoliteWhisper_SearchInProg = false; -- Class being searched on
PoliteWhisper_TaskQueue    = {};    -- Table of queued tasks
PoliteWhisper_Users        = {};    -- Table of users we've found/asked
PoliteWhisper_ChatUser     = nil;   -- Current user in chat frame.
PoliteWhisper_ListEntries  = 12;    -- 12 items on scroll list
PoliteWhisper_DoNotInvite  = {};

-- Reset everything once an hour
function PoliteWhisper_ResetAll()
    PoliteWhisper_SearchInProg = false;
    PoliteWhisper_TaskQueue    = {};
    PoliteWhisper_Users        = {};
    PoliteWhisper_HideChat(true);
end

-- PoliteWhisper_Invoke is called when the user types "/pw"
function PoliteWhisper_Invoke()
    local Class, i, Task, Reset;

    -- Update the ignore list
    PoliteWhisper_UpdateIgnore();

    -- Have we run before?
    if (PoliteWhisper_MinLevel:GetText() == "")
    then
        -- Set Min to be 4 levels below player
        L = UnitLevel("player");
        X = L - 4;
        if X < 1 then X = 1; end
        PoliteWhisper_MinLevel:SetText(X);

        -- Set Max to be 4 levels above player
        X = L + 4;
        if X > 70 then X = 70; end
        PoliteWhisper_MaxLevel:SetText(X);

        -- Set the defaults:
        -- Enable each class unless there is one present already in the party.
        -- I actually do this by enabling each class checkbox...
        for Class, Enable in pairs(PoliteWhisper_ClassEnables)
        do
            Enable:SetChecked(true);
        end
        -- ... then unchecking our class...
        PoliteWhisper_ClassEnables[UnitClass("player")]:SetChecked(false);
        -- ... then unchecking our group...
        for i = 1, GetNumPartyMembers()
        do
            PoliteWhisper_ClassEnables[UnitClass("party" ..
                i)]:SetChecked(false);
        end

        -- Perform searches no more than once per hour
        Task =
        {
            Expire = GetTime() + PoliteWhisper_ResetAllTimeout,
            Func   = PoliteWhisper_ResetAll
        };
        table.insert(PoliteWhisper_TaskQueue, Task);
    else
        -- Queue another ResetAll task?
        Reset = true;
        for i, Task in ipairs(PoliteWhisper_TaskQueue)
        do
            if Task.Func == PoliteWhisper_ResetAll
            then
                Reset = false;
                break;
            end
        end
        if (Reset)
        then
            Task =
            {
                Expire = GetTime() + PoliteWhisper_ResetAllTimeout,
                Func   = PoliteWhisper_ResetAll
            };
            table.insert(PoliteWhisper_TaskQueue, Task);
        end
    end

    -- Show the addon's frame
    PoliteWhisper_Frame:Show();

end

-- PoliteWhisper_OnLoad is called when the GUI is loaded
function PoliteWhisper_OnLoad()
    local i;

    -- Don't show the frames initially
    PoliteWhisper_Frame:Hide();

    -- Set tabs
    this.elapsed = 0;
    PanelTemplates_SetNumTabs(PoliteWhisper_Frame, 2);
    PanelTemplates_SetTab(PoliteWhisper_Frame, 1);

    -- Okay to drag frame.
    PoliteWhisper_Frame:RegisterForDrag("LeftButton");

    -- Bind slash command
    SlashCmdList["POLITE_WHISPER"] = PoliteWhisper_Invoke;
    SLASH_POLITE_WHISPER1 = "/pw";

    -- Tell user how to start program
    DEFAULT_CHAT_FRAME:AddMessage("Type \"/pw\" to open Polite Whisper");

    -- Initialize variables
    PoliteWhisper_ClassEnables =
    {
        Druid   = PoliteWhisper_Druid,
        Hunter  = PoliteWhisper_Hunter,
        Mage    = PoliteWhisper_Mage,
        Paladin = PoliteWhisper_Paladin,
        Priest  = PoliteWhisper_Priest,
        Rogue   = PoliteWhisper_Rogue,
        Shaman  = PoliteWhisper_Shaman,
        Warlock = PoliteWhisper_Warlock,
        Warrior = PoliteWhisper_Warrior
    };

    -- Locate the scrolling items (and hide the buttons)
    for i = 1, PoliteWhisper_ListEntries
    do
        Item =
        {
            Name  = _G["PoliteWhisper_Player" .. i],
            Class = _G["PoliteWhisper_Class" .. i],
            Level = _G["PoliteWhisper_Level" .. i],
            Loc   = _G["PoliteWhisper_Zone" .. i],
            Ask   = _G["PoliteWhisper_Ask" .. i]
        };
        table.insert(PoliteWhisper_ScrollList, Item);
        Item.Ask:Hide();
    end
    for i = 1, PoliteWhisper_History
    do
        table.insert(PoliteWhisper_ChatList, _G["PoliteWhisper_Chat" .. i]);
    end

    -- Hide chat pane.
    PoliteWhisper_HideChat();

    -- Configure scroll bars
    PoliteWhisper_ScrollBarText:SetText("");
    PoliteWhisper_ScrollBarLow:SetText("");
    PoliteWhisper_ScrollBarHigh:SetText("");
    PoliteWhisper_ScrollBar:SetValueStep(1);
	PoliteWhisper_ScrollBar:SetValue(0);
	PoliteWhisper_ScrollBar:Hide();
    PoliteWhisper_IgnoreBarText:SetText("");
    PoliteWhisper_IgnoreBarLow:SetText("");
    PoliteWhisper_IgnoreBarHigh:SetText("");
    PoliteWhisper_IgnoreBar:SetValueStep(1);
	PoliteWhisper_IgnoreBar:SetValue(0);
	PoliteWhisper_IgnoreBar:Hide();

    -- Catch /who events
    PoliteWhisper_Frame:RegisterEvent("WHO_LIST_UPDATE");
    PoliteWhisper_Frame:RegisterEvent("CHAT_MSG_WHISPER");
    PoliteWhisper_Frame:RegisterEvent("CHAT_MSG_AFK");
    PoliteWhisper_Frame:RegisterEvent("PARTY_MEMBERS_CHANGED");
    PoliteWhisper_Frame:RegisterEvent("VARIABLES_LOADED");
end

-- PoliteWhisper_Search is called when the user clicks the search button
function PoliteWhisper_Search()
    -- Fetch search parameters
    local Dngn = PoliteWhisper_Dungeon:GetText();
    local Min  = tonumber(PoliteWhisper_MinLevel:GetText());
    local Max  = tonumber(PoliteWhisper_MaxLevel:GetText());

    -- Validate search parameters
    if Dngn == ""
    then
        message("Invite people to join you doing what?");
        return;
    end

    if (not Min) or (Min > 70)
    then
        message("What's the minimum level character to invite?");
        return;
    end

    if (not Max) or (Max > 70)
    then
        message("What's the maximum level character to invite?");
        return;
    end

    -- Start search
    -- We don't actually want to loop through the searches at this time.
    -- Instead, we launch the first search and allow the search event to launch
    -- each successive search.
    PoliteWhisper_DoNextSearch();
end

-- PoliteWhisper_DoNextSearch is called to search for a given Class.
function PoliteWhisper_DoNextSearch(Class)
    local Enable, Min, Max, Filter;

    -- Find next class
    Class = next(PoliteWhisper_ClassEnables, Class)
    if not Class then return; end

    -- Search on this class?
    Enable = PoliteWhisper_ClassEnables[Class]
    if Enable:GetChecked()
    then
        -- Yes, start search
        Min  = tonumber(PoliteWhisper_MinLevel:GetText());
        Max  = tonumber(PoliteWhisper_MaxLevel:GetText());
        Filter = "c-\"" .. Class .. "\" " .. Min .. "-" .. Max;
        PoliteWhisper_SearchInProg = Class;
        SetWhoToUI(1);
        SendWho(Filter);

        -- Launch a timeout for the /who in case it doesn't complete.
        Task =
        {
            Expire = GetTime() + PoliteWhisper_WhoTimeout,
            Func   = PoliteWhisper_CancelWho
        };
        table.insert(PoliteWhisper_TaskQueue, Task);
    else
        -- No, skip to next
        PoliteWhisper_DoNextSearch(Class);
    end
end

-- Called if there is not a response to our /who.
function PoliteWhisper_CancelWho()
    -- Stop intercepting /who queries.
    PoliteWhisper_SearchInProg = false;
    SetWhoToUI(0);

    -- Tell the user that something went wrong.
    message("The server did not respond to our /who query.");
end

-- Called by PoliteWhisper_OnEvent in response to a WHO_LIST_UPDATE event.
function PoliteWhisper_OnWho()
    local i, Task, Class, Name, X, Level, PClass, Loc;

    -- Hide the window
    FriendsFrame:Hide();

    -- Remove the timeout event from the task queue.
    for i, Task in ipairs(PoliteWhisper_TaskQueue)
    do
        if Task.Func == PoliteWhisper_CancelWho
        then
            table.remove(PoliteWhisper_TaskQueue, i);
            break;
        end
    end

    -- We're done searching now.  If any more /who events fire, we'll ignore
    -- them.
    Class = PoliteWhisper_SearchInProg;
    PoliteWhisper_SearchInProg = false;
    SetWhoToUI(0);

    -- Save the results.
    for i = 1, GetNumWhoResults()
    do
        Name, X, Level, X, PClass, Loc, G = GetWhoInfo(i);
        PoliteWhisper_FoundWho(Name, Level, Class, Loc);
    end

    -- Continue search by scheduling another /who.
    Task =
    {
        Expire = GetTime() + PoliteWhisper_WhoDelay,
        Func   = PoliteWhisper_DoNextSearch,
        Arg1   = Class
    };
    table.insert(PoliteWhisper_TaskQueue, Task);

    -- Update display list.
    PoliteWhisper_UpdateList();
end

-- Called for each user returned by /who.
function PoliteWhisper_FoundWho(Name, Level, Class, Loc)
    local i, Status, Dungeon, L;

    -- Is the user on our ignore list?
    for i = 1, GetNumIgnores()
    do
        if Name == GetIgnoreName(i) then return; end
    end

    -- Is the user in our party?
    if Name == UnitName("player") then return; end
    for i = 1, GetNumPartyMembers()
    do
        if Name == UnitName("party" .. i) then return; end
    end

    -- Have we seen this user before?
    if PoliteWhisper_Users[Name]
    then
        Status = PoliteWhisper_Users[Name].Status;
    else
        -- Nope, create a record for them.
        PoliteWhisper_Users[Name] = { Level = Level, Class = Class };
    end

    -- Is the user in a dungeon?
    Dungeon = false;
    for i, L in ipairs(PoliteWhisper_Dungeons)
    do
        if Loc == L then Dungeon = true; end
    end

    -- What state should the user be in now?
    if (Status == "Re-Ask") and Dungeon
    then
        Status = "Dungeon";
    elseif (not Status) or (Status == "Ask") or (Status == "Dungeon") or
        (Status == "AFK")
    then
        if Dungeon
        then
            Status = "Dungeon";
        else
            Status = "Ask";
        end
    end

    -- Save the user's location and status
    PoliteWhisper_Users[Name].Loc    = Loc;
    PoliteWhisper_Users[Name].Status = Status;
end

-- Called in response to a captured event
function PoliteWhisper_OnEvent()
    local Msg, Name;

    -- Who event?
    if (event == "WHO_LIST_UPDATE") and PoliteWhisper_SearchInProg then
        PoliteWhisper_OnWho();
    elseif event == "CHAT_MSG_WHISPER" then
        PoliteWhisper_OnWhisper(arg2, arg1);
    elseif event == "CHAT_MSG_AFK" then
        PoliteWhisper_OnAFK(arg2);
    elseif event == "PARTY_MEMBERS_CHANGED" then
        PoliteWhisper_UpdateList();
    end
end

-- PoliteWhisper_OnUpdate is called periodically while the PW frame is visible.
-- We use this to launch any queued tasks.
function PoliteWhisper_OnUpdate()
    -- Loop through all queued tasks.
    local i, T;
    for i, T in ipairs(PoliteWhisper_TaskQueue)
    do
        -- Has task expired?
        if GetTime() > T.Expire
        then
            T.Func(T.Arg1, T.Arg2, T.Arg3);
            table.remove(PoliteWhisper_TaskQueue, i);
            return;
        end
    end
end

-- Clear out an item.
function PoliteWhisper_ShowItem(i, Name, Items)
    local Ctrls = PoliteWhisper_ScrollList[i];

    -- Show the user.
    Ctrls.Name:SetText(Name);
    Ctrls.Class:SetText(Items.Class);
    Ctrls.Level:SetText(Items.Level);
    Ctrls.Loc:SetText(Items.Loc);
    if (Name == PoliteWhisper_ChatUser)
    then
        Ctrls.Ask:SetText("Delete");
    else
        Ctrls.Ask:SetText(Items.Status);
    end
    Ctrls.Ask:Show();

    -- Remember the user associated with this entry for later.
    Ctrls.User = Name;
end

-- Returns true if this user's items match our filter.
function PoliteWhisper_Filter(Items)
    local Min, Max;
    Min = tonumber(PoliteWhisper_MinLevel:GetText());
    Max = tonumber(PoliteWhisper_MaxLevel:GetText());

    return (Items.Level >= Min) and (Items.Level <= Max) and
        PoliteWhisper_ClassEnables[Items.Class]:GetChecked();
end

-- Loop through a part of the user list. Called by PoliteWhisper_UpdateList.
function PoliteWhisper_ShowItems(Type, Offset)
    local Name, Items, ScrollOffset, i, InGroup;

    -- What is the scroll bar set to?
    ScrollOffset = PoliteWhisper_ScrollBar:GetValue();

    -- Loop through list
    for Name, Items in pairs(PoliteWhisper_Users)
    do
        -- Is this person in our group?
        InGroup = (Name == UnitName("player"));
        for i = 1, GetNumPartyMembers()
        do
            if UnitName("party" .. i) == Name then InGroup = true; end
        end

        -- In Ask and Re-Ask runs, we need to filter out users who no longer
        -- match our search criteria.
        if (not InGroup) and (Items.Status == Type) and
            (not PoliteWhisper_DoNotInvite[Name]) and
            ((Type == "Chat") or PoliteWhisper_Filter(Items))
        then
            Offset = Offset + 1;

            -- Show it?
            if (Offset > ScrollOffset) and
                (Offset <= (ScrollOffset + PoliteWhisper_ListEntries))
            then
                PoliteWhisper_ShowItem(Offset - ScrollOffset, Name, Items);
            end
        end
    end

    return Offset;
end

-- Clear out an item.
function PoliteWhisper_ClearItem(i)
    local Ctrls = PoliteWhisper_ScrollList[i];

    Ctrls.Name:SetText("");
    Ctrls.Class:SetText("");
    Ctrls.Level:SetText("");
    Ctrls.Loc:SetText("");
    Ctrls.Ask:Hide();
end

-- Update the ignore list display.
function PoliteWhisper_UpdateIgnore()
    local S, i, N, X;

    -- Loop through list
    i = 0;
    S = PoliteWhisper_IgnoreBar:GetValue();
    for N, X in pairs(PoliteWhisper_DoNotInvite)
    do
        i = i + 1;

        if ((i <= S) or ((i + S) > PoliteWhisper_IgnoreListEntries))
        then
            _G["PoliteWhisper_Ignore" .. (i - S)]:SetText("");
        else
            _G["PoliteWhisper_Ignore" .. (i - S)]:SetText(N);
        end
    end

    -- Is there a scroll bar?
    if (i > PoliteWhisper_IgnoreListEntries)
    then
        PoliteWhisper_IgnoreBar:SetMinMaxValues(0,
            i - PoliteWhisper_IgnoreListEntries);
        PoliteWhisper_IgnoreBar:SetValue(0);
    	PoliteWhisper_IgnoreBar:Show();
    else
    	PoliteWhisper_IgnoreBar:Hide();
    end

    -- Blank the rest of the list
    while (i < PoliteWhisper_IgnoreListEntries)
    do
        i = i + 1;
        _G["PoliteWhisper_Ignore" .. i]:SetText("");
    end
end

-- Format a name
function PoliteWhisper_FormatName(Name)
    return string.upper(string.sub(Name, 1, 1)) ..
        string.lower(string.sub(Name, 2));
end

-- Add a user to the ignore list
function PoliteWhisper_IgnoreAdd()
    local User = PoliteWhisper_FormatName(PoliteWhisper_Ignore:GetText());

    if (User == "")
    then
        message("Who should we add to the do not invite list?");
        return;
    end

    if (PoliteWhisper_DoNotInvite[User])
    then
        message("That player is already on the list.");
        return;
    end

    -- Update the display
    PlaySound("igCharacterInfoOpen");
    PoliteWhisper_DoNotInvite[User] = true;
    PoliteWhisper_UpdateIgnore();

    -- Clear out the text
    PoliteWhisper_Ignore:SetText("");

    -- Refilter
    PoliteWhisper_UpdateList();
end

-- Delete a user from the ignore list
function PoliteWhisper_IgnoreDelete()
    local User = PoliteWhisper_FormatName(PoliteWhisper_Ignore:GetText());

    if (User == "")
    then
        message("Who should we delete from the do not invite list?");
        return;
    end

    if (not PoliteWhisper_DoNotInvite[User])
    then
        message("That player is not on the list.");
        return;
    end

    -- Remove & update
    PoliteWhisper_DoNotInvite[User] = nil;
    PlaySound("igCharacterInfoClose");
    PoliteWhisper_Ignore:SetText("");
    PoliteWhisper_UpdateIgnore();

    -- Refilter
    PoliteWhisper_UpdateList();
end

-- Update the list display.  Called when PoliteWhisper_Users changes.
function PoliteWhisper_UpdateList()
    local i;

    -- Were we chatting with someone who is now a party member?
    for i = 1, GetNumPartyMembers()
    do
        if UnitName("party" .. i) == PoliteWhisper_ChatUser then
            -- Hide the chat pane.
            PoliteWhisper_HideChat();
        end
    end

    -- Show the entries.
    local i, Items;
    Items = PoliteWhisper_ShowItems("Chat", 0);
    Items = PoliteWhisper_ShowItems("Ask", Items);
    Items = PoliteWhisper_ShowItems("Re-Ask", Items);

    -- Clear any remaining items.
    for i = Items + 1, PoliteWhisper_ListEntries
    do
        PoliteWhisper_ClearItem(i);
    end

    -- Do we need a scroll bar?
    if Items > PoliteWhisper_ListEntries
    then
        -- Yup, make sure it is showing.
        PoliteWhisper_ScrollBar:Show();

        -- Adjust the limit and refresh.
        PoliteWhisper_ScrollBar:SetMinMaxValues(0,
            Items - PoliteWhisper_ListEntries);
        PoliteWhisper_ScrollBar:SetValue(PoliteWhisper_ScrollBar:GetValue());
    else
        -- Nope, make sure it is hidden.
        PoliteWhisper_ScrollBar:Hide();

        -- Were we scrolled any?
        if PoliteWhisper_ScrollBar:GetValue() > 0
        then
            -- Yes, crap, we have to redo the display.
            PoliteWhisper_ScrollBar:SetValue(0);
            PoliteWhisper_UpdateList();
        end
    end
end

-- Called when an ask button is clicked.
function PoliteWhisper_OnAsk(i)
    -- Find the user.
    local User = PoliteWhisper_ScrollList[i].User;

    -- Chat window already open with them?  That indicates "Delete".
    if (User == PoliteWhisper_ChatUser)
    then
        PoliteWhisper_Users[User].Status = "Deleted";
        PoliteWhisper_HideChat();
        return;
    end

    -- Open the chat window?
    if PoliteWhisper_Users[User].Status == "Chat"
    then
        -- Start chatting with the user.
        PoliteWhisper_ShowChat(User);
    else
        -- Must be some sort of "ask"
        Msg = string.format(PoliteWhisper_AskMessage,
            PoliteWhisper_Dungeon:GetText());
        SendChatMessage(Msg, "WHISPER", nil, User);

        -- Remove user from list. Create a chat log in case they reply.
        PoliteWhisper_Users[User].Status = "Pending";
        PoliteWhisper_Users[User].Log    = {"", "", "", ""};
        PoliteWhisper_UpdateList();

        -- It's okay to re-ask this user if they don't reply for 10 minutes.
        Task =
        {
            Expire = GetTime() + PoliteWhisper_NoRespPesterTimeout,
            Func   = PoliteWhisper_Repester,
            Arg1   = User
        };
        table.insert(PoliteWhisper_TaskQueue, Task);
    end
end

-- Called when it is okay to pester someone again.
function PoliteWhisper_Repester(User)
    -- They still in the list?
    if PoliteWhisper_Users[User]
    then
        -- Yup, add them in as Re-Ask.
        PoliteWhisper_Users[User].Status = "Re-Ask";
        PoliteWhisper_UpdateList();
    end
end

-- Called when a user is AFK.
function PoliteWhisper_OnAFK(Name)
    local M, U;

    -- Ignore messages from users we don't know.
    U = PoliteWhisper_Users[User];
    if not U then return; end

    -- Reset the pester timeout event to the shorter AFK time.
    for i, Task in ipairs(PoliteWhisper_TaskQueue)
    do
        if (Task.Func == PoliteWhisper_Repester) and (Task.Arg1 == User)
        then
            PoliteWhisper_TaskQueue[i].Expire = GetTime() +
                PoliteWhisper_AFKPesterTimeout;
            break;
        end
    end

    -- Update their status and display.
    if U.Status == "Pending"
    then
        U.Status = "AFK";
        PoliteWhisper_UpdateList();
    end
end

-- Called when a whisper is received.
function PoliteWhisper_OnWhisper(User, Msg)
    local M, U;

    if Msg == "QWERTYUIOP" then table.insert(PoliteWhisper_TaskQueue, 5); end

    -- Ignore messages from users we don't know.
    U = PoliteWhisper_Users[User];
    if not U then return; end

    -- Remove the pester timeout event from the task queue.
    for i, Task in ipairs(PoliteWhisper_TaskQueue)
    do
        if (Task.Func == PoliteWhisper_Repester) and (Task.Arg1 == User)
        then
            table.remove(PoliteWhisper_TaskQueue, i);
            break;
        end
    end

    -- Save the message.
    M   = string.format("[%s] whispers: %s", User, Msg);
    table.insert(U.Log, M)
    table.remove(U.Log, 1);

    -- Update their status and display.
    if (U.Status == "Pending") or (U.Status == "AFK")
    then
        U.Status = "Chat";
        PoliteWhisper_UpdateList();
    end

    -- Update the log.
    PoliteWhisper_UpdateLog();
end

-- Update the chat pane.
function PoliteWhisper_UpdateLog()
    local i, Log;

    -- Is the chat pane visible?
    if PoliteWhisper_ChatUser
    then
        Log = PoliteWhisper_Users[PoliteWhisper_ChatUser].Log;
        for i = 1, PoliteWhisper_History
        do
            PoliteWhisper_ChatList[i]:SetText(Log[i]);
        end
    end
end

-- Show the chat pane.
function PoliteWhisper_ShowChat(User)
    local i;

    -- Was it hidden?
    if not PoliteWhisper_ChatUser
    then
        -- Shrink scroll to minimum.
        PoliteWhisper_ListEntries = PoliteWhisper_MinListEntries;
        PoliteWhisper_UpdateList();
        PoliteWhisper_ScrollBar:SetHeight(PoliteWhisper_MinSliderHeight);
        for i = PoliteWhisper_MinListEntries + 1, PoliteWhisper_MaxListEntries
        do
            PoliteWhisper_ClearItem(i);
        end

        -- Feedback sound.
        PlaySound("igCharacterInfoOpen");
    end

    -- Save user.
    PoliteWhisper_ChatUser = User;

    -- Show stuff.
    PoliteWhisper_Talk:Show();
    PoliteWhisper_Decline:Show();
    PoliteWhisper_Invite:Show();
    PoliteWhisper_Party:Show();
    PoliteWhisper_UpdateLog();
    PoliteWhisper_UpdateList();
end

-- Hide the chat pane.
function PoliteWhisper_HideChat(Silent)
    local i;

    -- Not chatting with a user anymore.
    PoliteWhisper_ChatUser = nil;

    -- Hide stuff.
    PoliteWhisper_Talk:Hide();
    PoliteWhisper_Decline:Hide();
    PoliteWhisper_Invite:Hide();
    PoliteWhisper_Party:Hide();
    for i = 1, PoliteWhisper_History
    do
        PoliteWhisper_ChatList[i]:SetText("");
    end

    -- Expand scroll to maximum.
    PoliteWhisper_ListEntries = PoliteWhisper_MaxListEntries;
    PoliteWhisper_UpdateList();
    PoliteWhisper_ScrollBar:SetHeight(PoliteWhisper_MaxSliderHeight);

    -- Feedback sound.
    if (not Silent)
    then
        PlaySound("igCharacterInfoClose");
    end
end

-- User wants to send a whisper.
function PoliteWhisper_Speak()
    local Msg, Log;

    -- No action on blank.
    if (PoliteWhisper_Talk:GetText() == "") then return; end

    -- Log whisper.
    Msg = string.format("To [%s]: %s", PoliteWhisper_ChatUser,
        PoliteWhisper_Talk:GetText());
    Log = PoliteWhisper_Users[PoliteWhisper_ChatUser].Log;
    table.insert(Log, Msg)
    table.remove(Log, 1);

    -- Whisper user.
    SendChatMessage(PoliteWhisper_Talk:GetText(), "WHISPER", nil,
        PoliteWhisper_ChatUser);

    -- Update display.
    PoliteWhisper_UpdateLog();

    -- Remove old text.
    PoliteWhisper_Talk:SetText("");
end

-- User turned us down.
function PoliteWhisper_Declined()
    local User = PoliteWhisper_ChatUser;

    -- Close chat.
    PoliteWhisper_HideChat();

    -- Send user an "OK" message.
    SendChatMessage(PoliteWhisper_DeclineMessage, "WHISPER", nil, User);

    -- Remove user.
    PoliteWhisper_Users[User].Status = "Decline";
    PoliteWhisper_UpdateList();

end

-- Invite the user.
function PoliteWhisper_InviteUser()
    local Msg, Items, Leader;

    -- Are we the leader?
    if (GetNumPartyMembers() == 0) or IsPartyLeader("player")
    then
        InviteUnit(PoliteWhisper_ChatUser);
    else
        Items = PoliteWhisper_Users[PoliteWhisper_ChatUser];
        Msg = string.format(PoliteWhisper_InviteMessage, Items.Level,
            Items.Class, PoliteWhisper_ChatUser);
        Leader = UnitName("party" .. GetPartyLeaderIndex());
        SendChatMessage(Msg, "WHISPER", nil, Leader);
    end
end

-- Describe the current party.
function PoliteWhisper_TellParty()
    local Group, i, Msg;

    -- Create a message to describe our group.
    Group = PoliteWhisper_GroupText .. string.format(" a %s (%d)",
        UnitClass("player"), UnitLevel("player"))
    for i = 1, GetNumPartyMembers()
    do
        Group = Group .. string.format(", a %s (%d)",
            UnitClass("party" .. i), UnitLevel("party" .. i))
    end

    -- Add the message to their log.
    Msg = string.format("To [%s]: %s", PoliteWhisper_ChatUser, Group);
    Log = PoliteWhisper_Users[PoliteWhisper_ChatUser].Log;
    table.insert(Log, Msg)
    table.remove(Log, 1);

    -- Whisper user.
    SendChatMessage(Group, "WHISPER", nil, PoliteWhisper_ChatUser);

    -- Update display.
    PoliteWhisper_UpdateLog();
end
